# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from distutils.version import LooseVersion


MINIMAL_CLIENT_VERSION = LooseVersion('0.7.6')


# ALL THE FOLLOWING IMPORTS MOVED TO vault.py:
import xmlrpclib
from pylons import request
from base64 import b64encode
from datetime import *
from decorator import decorator
from pylons.controllers.xmlrpc import xmlrpc_fault

from sflvault.lib.base import *
from sflvault.lib.vault import SFLvaultAccess
from sflvault.model import *

from sqlalchemy import sql, exceptions

log = logging.getLogger(__name__)


#
# Permissions decorators for XML-RPC calls
#

def _authenticated_user_first(self, *args, **kwargs):
    """DRYed authenticated_user to skip repetition in authenticated_admin"""
    s = get_session(args[0])

    if not s:
        return vaultMsg(False, "Permission denied")

    self.sess = s

    if hasattr(self, 'vault'):
        if 'user_id' in self.sess:
            self.vault.myself_id = self.sess['user_id']
        if 'username' in self.sess:
            self.vault.myself_username = self.sess['username']


@decorator
def authenticated_user(func, self, *args, **kwargs):
    """Aborts if user isn't authenticated.

    Timeout check done in get_session.

    WARNING: authenticated_user READ the FIRST non-keyword argument
             (should be authtok)
    """
    ret = _authenticated_user_first(self, *args, **kwargs)
    if ret:
        return ret

    return func(self, *args, **kwargs)

@decorator
def authenticated_admin(func, self, *args, **kwargs):
    """Aborts if user isn't admin.

    Check authenticated_user , everything written then applies here as well.
    """
    ret = _authenticated_user_first(self, *args, **kwargs)
    if ret:
        return ret

    if not self.sess['userobj'].is_admin:
        return vaultMsg(False, "Permission denied, admin priv. required")

    return func(self, *args, **kwargs)


##
## See: http://wiki.pylonshq.com/display/pylonsdocs/Using+the+XMLRPCController
##
class XmlrpcController(XMLRPCController):
    """This controller is required to call model.Session.remove()
    after each call, otherwise, junk remains in the SQLAlchemy caches."""

    allow_none = True # Enable marshalling of None values through XMLRPC.
    
    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        self.vault = SFLvaultAccess()
        self.vault.setup_timeout = config['sflvault.vault.setup_timeout']

        try:
            return XMLRPCController.__call__(self, environ, start_response)
        # could be useful at some point:
        #except VaultError, e:
        #    print "ASLDJLASDJLASKJDKASJLD"
        #    return xmlrpc_fault(0, e.message)(environ, start_response)
        finally:
            model.meta.Session.remove()
    
    def sflvault_login(self, username, version=None):
        # Require minimal client version.        
        user_version = LooseVersion(version)
        if not version or user_version < MINIMAL_CLIENT_VERSION:
            return vaultMsg(False, "Minimal client version required: '%s'. "\
                            "You announced yourself as version '%s'" % \
                            (MINIMAL_CLIENT_VERSION.vstring, version))

        # Return 'cryptok', encrypted with pubkey.
        # Save decoded version to user's db field.
        try:
            u = query(User).filter_by(username=username).one()
        except Exception, e:
            return vaultMsg(False, "User unknown: %s" % e.message )
        
        # TODO: implement throttling ?

        rnd = randfunc(32)
        # 15 seconds to complete login/authenticate round-trip.
        u.logging_timeout = datetime.now() + timedelta(0, 15)
        u.logging_token = b64encode(rnd)

        meta.Session.flush()
        meta.Session.commit()

        e = u.elgamal()
        cryptok = serial_elgamal_msg(e.encrypt(rnd, randfunc(32)))
        return vaultMsg(True, 'Authenticate please', {'cryptok': cryptok})

    def sflvault_authenticate(self, username, cryptok):
        """Receive the *decrypted* cryptok, b64 encoded"""

        u = None
        try:
            u = query(User).filter_by(username=username).one()
        except:
            return vaultMsg(False, 'Invalid user')

        if u.logging_timeout < datetime.now():
            return vaultMsg(False, 'Login token expired. Now: %s Timeout: %s' % (datetime.now(), u.logging_timeout))

        # str() necessary, to convert buffer to string.
        if cryptok != str(u.logging_token):
            return vaultMsg(False, 'Authentication failed')
        else:
            newtok = b64encode(randfunc(32))
            set_session(newtok, {'username': username,
                                 'timeout': datetime.now() + timedelta(0, int(config['sflvault.vault.session_timeout'])),
                                 'remote_addr': request.environ.get('REMOTE_ADDR', None),
                                 'userobj': u,
                                 'user_id': u.id
                                 })

            return vaultMsg(True, 'Authentication successful', {'authtok': newtok})


    def sflvault_user_setup(self, username, pubkey):
        return self.vault.user_setup(username, pubkey)

    @authenticated_admin
    def sflvault_user_add(self, authtok, username, is_admin):
        return self.vault.user_add(username, is_admin)

    @authenticated_admin
    def sflvault_user_del(self, authtok, user):
        return self.vault.user_del(user)

    @authenticated_user
    def sflvault_user_list(self, authtok, groups=False):
        return self.vault.user_list(groups)

    @authenticated_user
    def sflvault_machine_get(self, authtok, machine_id):
        return self.vault.machine_get(machine_id)

    @authenticated_user
    def sflvault_machine_put(self, authtok, machine_id, data):
        return self.vault.machine_put(machine_id, data)

    @authenticated_user
    def sflvault_service_get(self, authtok, service_id, group_id=None):
        return self.vault.service_get(service_id, group_id)

    @authenticated_user
    def sflvault_service_get_tree(self, authtok, service_id, with_groups=False):
        return self.vault.service_get_tree(service_id, with_groups)

    @authenticated_user
    def sflvault_service_put(self, authtok, service_id, data):
        # 'user_id' required in session.
        # TODO: verify I had access to the service previously.
        req = sql.join(servicegroups_table, usergroups_table,
                       ServiceGroup.group_id==UserGroup.group_id) \
                 .join(users_table, User.id==UserGroup.user_id) \
                 .select() \
                 .where(User.id==self.sess['user_id']) \
                 .where(ServiceGroup.service_id==service_id)
        res = list(meta.Session.execute(req))
        if not res:
            return vaultMsg(False, "You don't have access to that service.")
        else:
            return self.vault.service_put(service_id, data)

    @authenticated_user
    def sflvault_search(self, authtok, search_query, group_ids=None,
                        verbose=False, filters=None):
        if group_ids and not filters:
            filters = {'groups': group_ids}
        if group_ids and isinstance(filters, dict) and 'groups' not in filters:
            # Please don't do that, use filters instead.
            filters['groups'] = group_ids
        return self.vault.search(search_query, filters, verbose)

    @authenticated_user
    def sflvault_service_add(self, authtok, machine_id, parent_service_id, url,
                             group_ids, secret, notes, metadata=None):
        return self.vault.service_add(machine_id, parent_service_id, url,
                                      group_ids, secret, notes, metadata)
        
    @authenticated_admin
    def sflvault_service_del(self, authtok, service_id):
        return self.vault.service_del(service_id)

    @authenticated_user
    def sflvault_service_list(self, authtok, machine_id=None, customer_id=None):
        return self.vault.service_list(machine_id, customer_id)

    @authenticated_user
    def sflvault_machine_add(self, authtok, customer_id, name, fqdn, ip,
                             location, notes):
        return self.vault.machine_add(customer_id, name, fqdn, ip,
                                      location, notes)

    @authenticated_admin
    def sflvault_machine_del(self, authtok, machine_id):
        return self.vault.machine_del(machine_id)

    @authenticated_user
    def sflvault_machine_list(self, authtok, customer_id=None):
        return self.vault.machine_list(customer_id)

    @authenticated_user
    def sflvault_customer_get(self, authtok, customer_id):
        return self.vault.customer_get(customer_id)

    @authenticated_user
    def sflvault_customer_put(self, authtok, customer_id, data):
        return self.vault.customer_put(customer_id, data)

    @authenticated_user
    def sflvault_customer_add(self, authtok, customer_name):
        return self.vault.customer_add(customer_name)

    @authenticated_admin
    def sflvault_customer_del(self, authtok, customer_id):
        return self.vault.customer_del(customer_id)

    @authenticated_user
    def sflvault_customer_list(self, authtok):
        return self.vault.customer_list()

    @authenticated_user
    def sflvault_group_get(self, authtok, group_id):
        return self.vault.group_get(group_id)

    @authenticated_user
    def sflvault_group_put(self, authtok, group_id, data):
        return self.vault.group_put(group_id, data)

    @authenticated_user
    def sflvault_group_add(self, authtok, group_name):
        return self.vault.group_add(group_name)

    @authenticated_admin
    def sflvault_group_del(self, authtok, group_id):
        return self.vault.group_del(group_id)

    @authenticated_user
    def sflvault_group_add_service(self, authtok, group_id, service_id, symkey):
        return self.vault.group_add_service(group_id, service_id, symkey)

    @authenticated_user
    def sflvault_group_del_service(self, authtok, group_id, service_id):
        fail = self._test_group_admin(group_id)
        if fail:
            return fail
                
        return self.vault.group_del_service(group_id, service_id)

    @authenticated_user
    def sflvault_group_add_user(self, authtok, group_id, user, is_admin=False,
                                cryptgroupkey=None):
        return self.vault.group_add_user(group_id, user, is_admin,
                                         cryptgroupkey)

    @authenticated_user
    def sflvault_group_del_user(self, authtok, group_id, user):
        fail = self._test_group_admin(group_id)
        if fail:
            return fail
        return self.vault.group_del_user(group_id, user)

    @authenticated_user
    def sflvault_group_list(self, authtok, list_users=False):
        return self.vault.group_list(False, list_users)

    @authenticated_user
    def sflvault_service_passwd(self, authtok, service_id, newsecret):
        return self.vault.service_passwd(service_id, newsecret)

    def _test_group_admin(self, group_id):
        if not query(Group).filter_by(id=group_id).first():
            return vaultMsg(False, "Group not found: %s" % str(e))

        # Verify if I'm is_admin on that group
        ug = query(UserGroup).filter_by(group_id=group_id,
                                        user_id=self.sess['user_id']).first()
        me = query(User).get(self.sess['user_id'])
        
        # Make sure I'm in that group (to be able to decrypt the groupkey)
        if not ug or (not ug.is_admin and not me.is_admin):
            return vaultMsg(False, "You are not admin on that group (nor global admin)")
        






# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The base Controller API

Provides the BaseController class for subclassing, and other objects
utilized by Controllers.
"""
from pylons import c, cache, config, g, request, response, session
from pylons.controllers import WSGIController, XMLRPCController
from pylons.controllers.util import abort, etag_cache, redirect_to
from pylons.decorators import jsonify, validate
from pylons.i18n import _, ungettext, N_
from pylons.templating import render
from datetime import *
import xmlrpclib

import sflvault.lib.helpers as h
from sflvault import model
from sflvault.model import query
from sflvault.common.crypto import *
from sflvault.common import VaultError

from base64 import b64decode, b64encode


# Helper to return messages

def vaultMsg(success, message, dict=None):
    """Form return message understandable by vault client"""
    ret = {'error': (not success), 'message': message}
    if dict:
        for x in dict:
            ret[x] = dict[x]
    return ret


#
# Session management functions
#
def _setup_sessions():
    """DRY out set_session and get_session"""
    if not hasattr(g, 'vaultSessions'):
        g.vaultSessions = {}

def set_session(authtok, value):
    """Sets in 'g.vaultSessions':
    {authtok1: {'username':  , 'timeout': datetime}, authtok2: {}..}
    
    """
    _setup_sessions();

    g.vaultSessions[authtok] = value;
        
def get_session(authtok):
    """Return the values associated with a session"""
    _setup_sessions();

    if not g.vaultSessions.has_key(authtok):
        return None

    if not g.vaultSessions[authtok].has_key('timeout'):
        g.vaultSessions[authtok]['timeout'] = datetime.now() + timedelta(0, SESSION_TIMEOUT)
    
    if g.vaultSessions[authtok]['timeout'] < datetime.now():
        del(g.vaultSessions[authtok])
        return None

    if g.vaultSessions[authtok]['remote_addr'] != request.environ.get('REMOTE_ADDR', 'gibberish'):
        del(g.vaultSessions[authtok])
        return None

    return g.vaultSessions[authtok]



class BaseController(WSGIController):
    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # WSGIController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']
        try:
            return WSGIController.__call__(self, environ, start_response)
        finally:
            model.meta.Session.remove()

# XMLRPC definition bac in xmlrpc.py


# Include the '_' function in the public names
__all__ = [__name for __name in locals().keys() if not __name.startswith('_') \
           or __name == '_']



# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from base64 import b64decode, b64encode
from datetime import datetime
import re

from Crypto.PublicKey import ElGamal
from sqlalchemy import Column, MetaData, Table, types, ForeignKey
from sqlalchemy.orm import mapper, relation, backref
from sqlalchemy.orm import scoped_session, sessionmaker, eagerload, lazyload
from sqlalchemy.orm import eagerload_all
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy import sql

from sflvault.model import meta
from sflvault.model.meta import Session, metadata
from sflvault.model.custom_types import JSONEncodedDict
from sflvault.common.crypto import *


# TODO: add an __all__ statement here, to speed up loading...


def init_model(engine):
    """Call me before using any of the tables or classes in the model."""
    sm = sessionmaker(autoflush=True, transactional=True, bind=engine)

    meta.engine = engine
    meta.Session = scoped_session(sm)



users_table = Table("users", metadata,
                    Column('id', types.Integer, primary_key=True),
                    Column('username', types.Unicode(50)),
                    # ElGamal user's public key.
                    Column('pubkey', types.Text),
                    # Used in the login/authenticate challenge
                    Column('logging_token', types.Binary(35)),
                    # Time until the token is valid.
                    Column('logging_timeout', types.DateTime),
                    # This stamp is used to wipe users which haven't 'setup'
                    # their account before this date/time
                    Column('waiting_setup', types.DateTime, nullable=True),
                    Column('created_time', types.DateTime,
                           default=datetime.now()),
                    # Admin flag, allows to add users, and grant access.
                    Column('is_admin', types.Boolean, default=False)
                    )

usergroups_table = Table('users_groups', metadata,
                         Column('id', types.Integer, primary_key=True),
                         Column('user_id', types.Integer,
                                ForeignKey('users.id')),
                         Column('group_id', types.Integer,
                                ForeignKey('groups.id')),
                         Column('is_admin', types.Boolean, default=False),
                         Column('cryptgroupkey', types.Text),
                         )

groups_table = Table('groups', metadata,
                     Column('id', types.Integer, primary_key=True),
                     Column('name', types.Unicode(50)),
                     Column('hidden', types.Boolean, default=False),
                     # ElGamal group's public key
                     Column('pubkey', types.Text),
                     )


servicegroups_table = Table('services_groups', metadata,
                            Column('id', types.Integer, primary_key=True),
                            Column('service_id', types.Integer,
                                   ForeignKey('services.id')),
                            Column('group_id', types.Integer,
                                   ForeignKey('groups.id')),
                            Column('cryptsymkey', types.Text),
                            )


customers_table = Table('customers', metadata,
                        Column('id', types.Integer, primary_key=True),
                        Column('name', types.Unicode(100)),
                        Column('created_time', types.DateTime),
                        # username, même si yé effacé.
                        Column('created_user', types.Unicode(50))
                        )

machines_table = Table('machines', metadata,
                      Column('id', types.Integer, primary_key=True),
                      Column('customer_id', types.Integer, ForeignKey('customers.id')), # relation customers
                      Column('created_time', types.DateTime),
                      # Unicode lisible, un peu de descriptif
                      Column('name', types.Unicode(150)),
                      # Domaine complet.
                      Column('fqdn', types.Unicode(150)),
                      # Adresse IP si fixe, sinon 'dyn'
                      Column('ip', types.String(100)),
                      # Où il est ce serveur, location géographique, et dans
                      # la ville et dans son boîtier (4ième ?)
                      Column('location', types.Text),
                      # Notes sur le serveur, références, URLs, etc..
                      Column('notes', types.Text)
                      )

# Each ssh or web app. service that have a password.
services_table = Table('services', metadata,
                       Column('id', types.Integer, primary_key=True),
                       # Service lies on which Machine ?
                       Column('machine_id', types.Integer,
                              ForeignKey('machines.id')),
                       # Hierarchical service required to access this one ?
                       Column('parent_service_id', types.Integer,
                              ForeignKey('services.id')),
                       # REMOVED: replaced by servicegroups_table many-to-many.
                       #Column('group_id', types.Integer,
                       #       ForeignKey('groups.id')),
                       Column('url', types.String(250)), # Full service desc.
                       # simplejson'd python structures, depends on url scheme
                       Column('metadata', JSONEncodedDict), # reserved.
                       Column('notes', types.Text),
                       Column('secret', types.Text),
                       Column('secret_last_modified', types.DateTime)
                       )


class Service(object):
    def __repr__(self):
        return "<Service s#%d: %s>" % (self.id, self.url)

class Machine(object):
    def __repr__(self):
        return "<Machine m#%d: %s (%s %s)>" % (self.id if self.id else 0,
                                               self.name, self.fqdn, self.ip)
class User(object):
    def setup_expired(self):
        """Return True/False if waiting_setup has expired"""
        if self.waiting_setup and self.waiting_setup < datetime.now():
            return True
        else:
            return False

    def elgamal(self):
        """Return the ElGamal object, ready to encrypt stuff."""
        e = ElGamal.ElGamalobj()
        (e.p, e.g, e.y) = unserial_elgamal_pubkey(self.pubkey)
        return e
    
    def __repr__(self):
        return "<User u#%d: %s>" % (self.id, self.username)

class UserGroup(object):
    """Membership of a user to a group"""
    def __init__(self, user=None):
        if user:
            self.user = user

    def __repr__(self):
        return "<UserGroup element>"

class ServiceGroup(object):
    """membership of a service to a group"""
    def __init__(self, service=None):
        if service:
            self.service = service
        
    def __repr__(self):
        return "<ServiceGroup element>"

class Group(object):
    def __repr__(self):
        return "<Group: %s>" % (self.name)
    
    def elgamal(self):
        """Return the ElGamal object, ready to encrypt stuff."""
        e = ElGamal.ElGamalobj()
        (e.p, e.g, e.y) = unserial_elgamal_pubkey(self.pubkey)
        return e

class Customer(object):
    def __repr__(self):
        return "<Customer c#%d: %s>" % (self.id, self.name)

# User
#  .groups_assoc
#    UserGroup
#     .group
#       Group
#        .services_assoc
#          ServiceGroup
#           .service
#             Service
# Service
#  .groups_assoc
#    ServiceGroup
#     .group
#       Group
#        .users_assoc
#          UserGroup
#           .user
#             User

# Map each class to its corresponding table.
mapper(User, users_table, {
    # Quick access to services...
    'services': relation(Service,
                         secondary=usergroups_table.join(servicegroups_table, usergroups_table.c.group_id==servicegroups_table.c.group_id),
                         backref='users',
                         viewonly=True,
                         ),
    'groups_assoc': relation(UserGroup, backref='user')
    })
User.groups = association_proxy('groups_assoc', 'group')

mapper(UserGroup, usergroups_table, {
    'group': relation(Group, backref='users_assoc')
    })

mapper(Group, groups_table, {
    'services_assoc': relation(ServiceGroup, backref='group')
    })
Group.users = association_proxy('users_assoc', 'user')
Group.services = association_proxy('services_assoc', 'service')

mapper(ServiceGroup, servicegroups_table, {
    'service': relation(Service, backref='groups_assoc')
    })

mapper(Service, services_table, {
    'children': relation(Service,
                         lazy=False,
                         backref=backref('parent', uselist=False,
                                         remote_side=[services_table.c.id]),
                         primaryjoin=services_table.c.parent_service_id==services_table.c.id)
    })
Service.groups = association_proxy('groups_assoc', 'group')

mapper(Machine, machines_table, {
    'services': relation(Service, backref='machine', lazy=False)
    })
mapper(Customer, customers_table, {
    'machines': relation(Machine, backref='customer', lazy=False)
    })

################ Helper functions ################

def query(cls):
    """Shortcut to meta.Session.query(cls)"""
    return meta.Session.query(cls)


def get_user(user, eagerload_all_=None):
    """Get a user provided a username or an int(user_id), possibly eager
    loading some relations.
    """
    if isinstance(user, int):
        uq = query(User).filter_by(id=user)
    else:
        uq = query(User).filter_by(username=user)

    if eagerload_all_:
        uq = uq.options(eagerload_all(eagerload_all_))

    usr = uq.first()
    
    if not usr:
        raise LookupError("Invalid user: %s" % user)

    return usr


def get_objects_ids(objects_ids, object_type):
    """Return a list of valid IDs for certain object types.

    objects_ids - Must be a list of str or ints
    object_type - One of 'groups', 'machines', 'customers
    """
    return get_objects_list(objects_ids, object_type, return_objects=False)[1]


def get_objects_list(objects_ids, object_type, eagerload_all_=None,
                     return_objects=True):
    """Get a list of objects by their IDs, either as int or str. Make
    sure we return a list of integers as IDs.

    object_type - the type of object to be returned. It must be one of
                ['groups', 'machines', 'customers']
    return_objects - whether to return the actual objects or not.
    """

    objects_types_assoc = {'groups': Group,
                           'machines': Machine,
                           'customers': Customer}

    # Check if object_type is valid
    if object_type not in objects_types_assoc:
        raise ValueError("Invalid object type: %s" % (object_type))

    # Get variables
    if isinstance(objects_ids, str):
        objects_ids = [int(objects_ids)]
    elif isinstance(objects_ids, int):
        objects_ids = [objects_ids]
    elif isinstance(objects_ids, list):
        objects_ids = [int(x) for x in objects_ids]
    else:
        raise ValueError("Invalid %s specification" % (object_type))

    # Pull the objects/IDs from the DB
    obj = objects_types_assoc[object_type]
    if return_objects:
        objects_q = query(obj).filter(obj.id.in_(objects_ids))

        if eagerload_all_:
            objects_q = objects_q.options(eagerload_all(eagerload_all_))
        objects = objects_q.all()
    else:
        objects_q = sql.select([obj.id]).where(obj.id.in_(objects_ids))
        objects = meta.Session.execute(objects_q).fetchall()
        
    if len(objects) != len(objects_ids):
        # Woah, you specified objects that didn't exist ?
        valid_objects = [x.id for x in objects]
        invalid_objects = [x for x in objects_ids if x not in valid_objects]
        raise ValueError("Invalid %s: %s" % (object_type, invalid_objects))

    return (objects if return_objects else None, objects_ids)


def search_query(swords, filters=None, verbose=False):

    # Create the join..
    sel = sql.outerjoin(customers_table, machines_table).outerjoin(services_table)

    if filters:
        # Remove filters that are just None
        filters = dict([(x, filters[x]) for x in filters if filters[x]])

        if not isinstance(filters, dict):
            raise RuntimeError("filters param must be a dict, or None")

        if [True for x in filters if not isinstance(filters[x], list)]:
            raise RuntimeError("filters themselves must be a list of ints")
        
        if 'groups' in filters:
            sel = sel.join(servicegroups_table)

    sel = sel.select(use_labels=True)

    if filters:
        if 'groups' in filters:
            sel = sel.where(ServiceGroup.group_id.in_(filters['groups']))
        if 'machines' in filters:
            sel = sel.where(Machine.id.in_(filters['machines']))
        if 'customers' in filters:
            sel = sel.where(Customer.id.in_(filters['customers']))

    # Fields to search in..
    textfields = [Customer.name,
                  Machine.name,
                  Machine.fqdn,
                  Machine.ip,
                  Machine.location,
                  Machine.notes,
                  Service.url,
                  Service.notes]
    numfields = [Customer.id,
                 Machine.id,
                 Service.id]
    
    # TODO: distinguish between INTEGER fields and STRINGS and search
    # differently (check only ==, and only if word can be converted to int())

    andlist = []
    for word in swords:
        orlist = [field.ilike('%%%s%%' % word) for field in textfields]
        if word.isdigit():
            # Search numeric fields too
            orlist += [field == int(word) for field in numfields]
        orword = sql.or_(*orlist)
        andlist.append(orword)

    sel = sel.where(sql.and_(*andlist))

    sel = sel.order_by(Machine.name, Service.url)

    return meta.Session.execute(sel)




"""SQLAlchemy Metadata and Session object"""
from sqlalchemy import MetaData

__all__ = ['engine', 'metadata', 'Session']

# SQLAlchemy database engine.  Updated by model.init_model().
engine = None

# SQLAlchemy session manager.  Updated by model.init_model().
Session = None

# Global metadata. If you have multiple databases with overlapping table 
# names, you'll need a metadata for each database.
metadata = MetaData()

# For use in paster shell:
# from sqlalchemy import create_engine
# model.init_model(create_engine('sqlite:///./vault-database.db', echo=True))



"""Setup the SFLvault application"""
import logging

from paste.deploy import appconfig
from pylons import config
from datetime import datetime, timedelta

from sflvault.config.environment import load_environment

log = logging.getLogger(__name__)

def setup_config(command, filename, section, vars):
    """Place any commands to setup sflvault here"""
    conf = appconfig('config:' + filename)
    load_environment(conf.global_conf, conf.local_conf)

    from sflvault import model
    log.info("Creating tables")
    model.meta.metadata.create_all(bind=model.meta.engine)
    ## Add default user 'admin' with 5 minutes for the client to call "sflvault setup"
    if model.query(model.User).filter_by(username='admin').first():
        log.info("User 'admin' already exists, skipping insertion of admin user.")
    else:
        u = model.User()
        u.waiting_setup = datetime.now() + timedelta(0, 900)
        u.username = u'admin'
        u.created_time = datetime.now()
        u.is_admin = True

        model.meta.Session.add(u)        
        model.meta.Session.commit()

        log.info("Added 'admin' user, you have 15 minutes to setup from your client")
    
    log.info("Successfully setup")



import simplejson

from sqlalchemy import types

class JSONEncodedDict(types.TypeDecorator):
    """Represents an mutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = types.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = simplejson.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        return simplejson.loads(value) if value else {}

    def copy_value(self, value):
        return simplejson.loads(simplejson.dumps(value))



# -=- encoding: utf-8 -=-
from sflvault.tests import TestController
from sflvault.common.crypto import *
from sflvault.client.client import authenticate

import logging

log = logging.getLogger('tester')

class TestVaultController(TestController):
    
    def setUp(self):
        self.vault = self.getVault()

    def test_customer_add(self):
        """testing add a new customer to the vault"""
        res = self.vault.customer_add('testing customer 1 2 3')
        cid1 = res['customer_id']
        self.assertTrue(cid1, 1)
        self.assertEqual(res['message'], 'Customer added')

    def test_machine_add(self):
        """testing add a new machine to the vault"""
        res = self.vault.customer_add(u"Testing é les autres")
        cid2 = res['customer_id']
        self.assertEqual(res['message'], 'Customer added')    
        res = self.vault.machine_add(str(cid2), 
                                     "Machine name 2",
                                     "domain1.example2.com", 
                                     '4.3.2.1',
                                     None, 
                                     None)
        self.assertFalse("Error adding machine" in res['message'])

    def test_service_add(self):
        """testing add a new service to the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        mres = self.vault.machine_add(str(cres['customer_id']), 
                                      "Machine name 3",
                                      "domain1.example2.com", 
                                      '4.3.2.1',
                                      None, 
                                      None)
        res = self.vault.service_add(cres['customer_id'], 
                                     mres['machine_id'],
                                     0, 
                                     'ssh://marrakis@localhost',
                                     [], 
                                     'test',
                                     '')
        self.assertFalse("Error adding service" in res['message'])

    def test_alias_add(self):
        """testing add a new alias to the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        ares = self.vault.alias_add("#%s"%cres['customer_id'], 'c#1')
        self.assertTrue(ares)

    def test_user_add(self):
        """testing add a new user to the vault"""
        ures = self.vault.user_add("test_username")
        self.assertTrue("User added" in ures['message'])
        ures = self.vault.user_add("test_admin", True)
        self.assertTrue("Admin user added" in ures['message'])

    def test_group_add(self):
        """testing add a new group to the vault"""
        gres = self.vault.group_add("test_group")
        self.assertTrue(int(gres['group_id']) > 0)

    def test_group_add_user(self):
        """testing add a user to a group to the vault"""
        gres1 = self.vault.group_add("test_group1_user")
        gres2 = self.vault.group_add("test_group2_user")

        gares1 = self.vault.group_add_user(gres1['group_id'], 1)
        gares2 = self.vault.group_add_user(gres2['group_id'], 1, True)

        self.assertFalse("Error adding user to group" in gares1)
        self.assertFalse("Error adding user to group" in gares2)

    def test_group_add_service(self):
        """testing add a service to a group to the vault"""
        gres3 = self.vault.group_add("test_group3_user")
        cres = self.vault.customer_add(u"Testing é les autres")
        gares1 = self.vault.group_add_user(gres3['group_id'], 1)
        mres = self.vault.machine_add(str(cres['customer_id']), 
                                      "Machine name 3",
                                      "domain1.example2.com", 
                                      '4.3.2.1',
                                      None, 
                                      None)
        res = self.vault.service_add(cres['customer_id'], 
                                     mres['machine_id'],
                                     0, 
                                     'ssh://marrakis@localhost',
                                     [], 
                                     'test',
                                     '')
        gares3 = self.vault.group_add_service(gres3['group_id'], res['service_id'])

        self.assertFalse("Error adding service to group" in gares3)

    def test_customer_del(self):
        """testing delete a new customer from the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        dres = self.vault.customer_del(cres['customer_id'])
        self.assertTrue(dres is not None)

    def test_machine_del(self):
        """testing delete a machine from the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        mres = self.vault.machine_add(str(cres['customer_id']), "Machine name 3",
                                      "domain1.example2.com", '4.3.2.1',
                                      None, None)
        self.assertTrue(mres is not None)

    def test_service_del(self):
        """testing delete a service from the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        mres = self.vault.machine_add(str(cres['customer_id']), 
                                      "Machine name 3",
                                      "domain1.example2.com", 
                                      '4.3.2.1',
                                      None, 
                                      None)
        sres = self.vault.service_add(cres['customer_id'], 
                                      mres['machine_id'],
                                      0, 
                                      'ssh://marrakis@localhost',
                                      [], 
                                      'test',
                                      '')
        sres = self.vault.service_del(sres['service_id'])
        self.assertTrue(sres is not None)

        
    def test_alias_del(self):
        """testing delete an alias from the vault"""
        cres = self.vault.customer_add(u"Testing é les autres")
        ares = self.vault.alias_add("#%s"%cres['customer_id'], 'c#23')
        dres = self.vault.alias_del('c#23')
        self.assertTrue(dres is not None)

    def test_user_del(self):
        """testing delete a user from the vault"""
        self.assertTrue(False)

    def test_goup_del(self):
        """testing delete a group from the vault"""
        self.assertTrue(False)

    def test_goup_del_user(self):
        """testing delete a user from a group from the vault"""
        self.assertTrue(False)

    def test_goup_del_service(self):
        """testing delete a service from a group from the vault"""
        self.assertTrue(False)



"""Pylons application test package

When the test runner finds and executes tests within this directory,
this file will be loaded to setup the test environment.

It registers the root directory of the project in sys.path and
pkg_resources, in case the project hasn't been installed with
setuptools. It also initializes the application via websetup (paster
setup-app) with the project's test.ini configuration file.
"""
import os
import sys
from unittest import TestCase

import pkg_resources
import paste.fixture
import paste.script.appinstall
from paste.deploy import loadapp
from routes import url_for
from ConfigParser import ConfigParser
from paste.httpserver import serve
import threading

from pylons import config
from sflvault.client import SFLvaultClient
from sflvault.lib.vault import SFLvaultAccess

__all__ = ['url_for', 'TestController', 'setUp', 'tearDown']

here_dir = os.path.dirname(os.path.abspath(__file__))
conf_dir = os.path.dirname(os.path.dirname(here_dir))

sys.path.insert(0, conf_dir)
pkg_resources.working_set.add_entry(conf_dir)
pkg_resources.require('Paste')
pkg_resources.require('PasteScript')

dbfile = os.path.join(conf_dir, 'test-database.db')
confile = os.path.join(conf_dir, 'test-config')
test_file = os.path.join(conf_dir, 'test.ini')
globs = {}
vault = None


def tearDown():
    """Close the SFLVault server"""
    globs['server'].server_close()


def setUp():
    """Setup the temporary SFLvault server"""
    # Remove the test database on each run.
    if os.path.exists(dbfile):
        os.unlink(dbfile)

    # Remove the test config on each run
    if os.path.exists(confile):
        os.unlink(confile)

    cmd = paste.script.appinstall.SetupCommand('setup-app')
    cmd.run([test_file])

    cfg = ConfigParser()
    cfg.read(test_file)
    sinfos = cfg._sections['server:main']
    wsgiapp = loadapp('config:test.ini', relative_to=conf_dir)
    server = serve(wsgiapp, sinfos['host'], sinfos['port'],
                   socket_timeout=1, start_loop=False)
    globs['server'] = server
    t = threading.Thread(target=server.serve_forever)
    t.setDaemon(True)
    t.start()

    wsgiapp = loadapp('config:test.ini', relative_to=conf_dir)
    app = paste.fixture.TestApp(wsgiapp)

    # Create the vault test obj.
    if 'SFLVAULT_ASKPASS' in os.environ:
        del(os.environ['SFLVAULT_ASKPASS'])
    os.environ['SFLVAULT_CONFIG'] = config['sflvault.testconfig']        



class TestController(TestCase):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)

    def getVault(self):
        """Get the SFLVault server vault"""
        import sflvault.tests
        if sflvault.tests.vault is None:
            sflvault.tests.vault = SFLvaultClient(os.environ['SFLVAULT_CONFIG'],
                                                  shell=True)
            sflvault.tests.vault.passphrase = 'test'
            sflvault.tests.vault.username = 'admin'    

            def givepass():
                return sflvault.tests.vault.passphrase        
            sflvault.tests.vault.set_getpassfunc(givepass)
            sflvault.tests.vault.user_setup(sflvault.tests.vault.username, 
                                            'http://localhost:5551/vault/rpc', 
                                            sflvault.tests.vault.passphrase)
            self.cfg = sflvault.tests.vault.cfg
        return sflvault.tests.vault
    



# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SFLvaultAccess class file: main Vault access object.

This object is used over XML-RPC for remote requests, and used directly
for local access to the vault.

It is assumed here that full access is granted, or that security
filtering has been done at an earlier level (via an XML-RPC mechanism
for example).

This class only provides all the abstraction to the Vault, and should
return clean responses to clean queries.
"""


import xmlrpclib
from base64 import b64decode, b64encode
from datetime import *
import time as stdtime

from sqlalchemy import sql
from sqlalchemy.exceptions import InvalidRequestError as InvalidReq
from sqlalchemy.orm import eagerload_all

from sflvault.model import *
from sflvault.lib.base import *

import logging
log = logging.getLogger('sflvault')



class SFLvaultAccess(object):
    def __init__(self):
        """Init obj."""

        # This gives this object the knowledge of which user_id is currently
        # using the Vault
        self.myself_id = None
        # This gives this object the knowledge of which username is currently
        # using the Vault
        self.myself_username = None

        self.setup_timeout = 300

    def _log_any(self, log_func, msg, data):
        # Need to do that for user-setup
        if self.myself_username == None and self.myself_id == None:
            head = "NO USER - "
        else:
            head = "User: u#d%d - %s - " % (self.myself_id, self.myself_username)
        log_func(head + msg % data)

    def log_e(self, msg, data=None):
        self._log_any(log.error, msg, data)

    def log_i(self, msg, data=None):
        self._log_any(log.info, msg, data)

    def log_w(self, msg, data=None):
        self._log_any(log.warning, msg, data)


    def user_setup(self, username, pubkey):
        """Setup the user's account"""
        from pylons import app_globals

        u = query(User).filter_by(username=username).first()

        if u is None:
            self.log_i('No such temporary user %(username)s, add it first',
                       {"username": username})
            return vaultMsg(False, 'No such temporary user %s, add it first' % username)
        
        if u.setup_expired():
            self.log_i('Setup expired for user %(username)s',
                       {"username": username})
            return vaultMsg(False, 'Setup expired for user %s' % username)

        if u.pubkey:
            self.log_i('User %(username)s already has a publics',
                       {"username": username})
            return vaultMsg(False, 'User %s already has a public ' \
                                'key stored' % username)

        # Ok, let's save the things and reset waiting_setup.
        u.waiting_setup = None
        u.pubkey = pubkey
        
        # Save new informations
        meta.Session.add(u)
        meta.Session.commit()

        self.log_i('User setup complete for %(username)s',
                   {"username": username})
        return vaultMsg(True, 'User setup complete for %s' % username)


    def user_add(self, username, is_admin):
        usr = query(User).filter_by(username=username).first()

        msg = ''

        if usr is None:
            # New user
            usr = User()
            usr.waiting_setup =  datetime.now() + timedelta(0, int(self.setup_timeout))
            usr.username = username
            usr.is_admin = bool(is_admin)
            usr.created_time = datetime.now()

            meta.Session.add(usr)
            
            self.log_i('User %(username)s added.', {"username": username})
            msg = 'added'
        elif usr.waiting_setup:
            if usr.waiting_setup < datetime.now():
                # Verify if it's a waiting_setup user that has expired.
                usr.waiting_setup = datetime.now() + \
                                    timedelta(0, int(self.setup_timeout))
            
                self.log_i('updated (had setup timeout expired).', {})
                msg = 'updated (had setup timeout expired)'
            else:
                self.log_i('User %(username)s is waiting for setup.',
                       {"username": username})
                return vaultMsg(False, "User %s is waiting for setup" % \
                                username)
        else:
            self.log_i('User %(username)s already exists.',
                       {"username": username})
            return vaultMsg(False, 'User %s already exists.' % username)

        meta.Session.commit()

        return vaultMsg(True, '%s %s. User has a delay of %d seconds '
                              'to invoke a "user-setup" command' % \
                            ('Admin user' if is_admin else 'User',
                             msg, int(self.setup_timeout)), {'user_id': usr.id})
        

    def user_del(self, user):
        """Delete a user from database.

        :param user: can be a user_id, or username

        This will remove the user and it's links to groups.
        
        """
        try:
            usr = model.get_user(user)
        except LookupError, e:
            self.log_w('User UNsuccessfully deleted: %(error)s',
                        {'error': str(e)})
            return vaultMsg(False, str(e))

        t1 = model.usergroups_table
        meta.Session.execute(t1.delete(t1.c.user_id==usr.id))
        username = usr.username
        meta.Session.delete(usr)
        meta.Session.commit()
        self.log_i('User %(username)s successfully deleted.',
                   {"username": username})
        return vaultMsg(True, "User %s successfully deleted" % username)


    def user_list(self, groups=False):
        """Return a simple list of the users

        groups - return the list of groups for each user, or not
        """
        req = query(User)
        if groups:
            req = req.options(eagerload_all('groups_assoc.group'))
        lst = query(User).all()

        out = []
        for x in lst:
            # perhaps add the pubkey ?
            if x.created_time:
                stmp = xmlrpclib.DateTime(x.created_time)
            else:
                stmp = 0
                
            nx = {'id': x.id, 'username': x.username,
                  'created_stamp': stmp,
                  'is_admin': x.is_admin,
                  'setup_expired': x.setup_expired(),
                  'waiting_setup': bool(x.waiting_setup)}

            if groups:
                # Loop group_assocs for that user..
                nx['groups'] = []
                for ug in x.groups_assoc:
                    nxgrp = {'is_admin': ug.is_admin,
                             'name': ug.group.name,
                             'id': ug.group_id}
                    nx['groups'].append(nxgrp)
                    
            out.append(nx)

        # Can use: datetime.fromtimestamp(x.created_stamp)
        # to get a datetime object back from the x.created_time
        return vaultMsg(True, "Here is the user list", {'list': out})
        

    def service_get(self, service_id, group_id=None):
        """Get a single service's data.

        group_id - return this group's key, otherwise, use first available
        """
        try:
            out = self._service_get_data(service_id, group_id)
        except VaultError, e:
            return vaultMsg(False, str(e))

        return vaultMsg(True, "Here is the service", {'service': out})
    

    def service_put(self, service_id, data):
        """Put a single service's data back to the vault's database"""

        try:
            s = query(Service).filter_by(id=service_id).one()
        except InvalidReq, e:
            self.log_i('Service not found: %(error)s', {"error": str(e)})
            return vaultMsg(False, "Service not found: %s" % str(e))

        #Save:
        # machine_id
        # parent_service_id
        # url
        # notes
        # location
        if 'machine_id' in data:
            s.machine_id = int(data['machine_id'])
        if 'parent_service_id' in data:
            if data['parent_service_id'] == '':
                s.parent_service_id = None
            else:
                s.parent_service_id = int(data['parent_service_id'])
        if 'notes' in data:
            s.notes = data['notes']
        if 'url' in data:
            s.url = data['url']
        if 'metadata' in data:
            s.metadata = data['metadata']

        meta.Session.commit()

        self.log_i('Service s#(service_id)s saved successfully' ,
                   {"service_id": service_id})
        return vaultMsg(True, "Service s#%s saved successfully" % service_id)

    def _service_get_data(self, service_id, group_id=None, with_groups=False):
        """Retrieve the information for a given service."""
        try:
            s = query(Service).filter_by(id=service_id).one()
        except InvalidReq, e:
            self.log_w('Service not found: %(service_id)s (%(error)s)',
                       {"service_id": service_id, "error": str(e)})
            raise VaultError("Service not found: %s (%s)" % (service_id,
                                                             str(e)))
        # unused
        #me = query(User).get(self.myself_id)

        # We need no aliasing, because we'll only use `cryptgroupkey`,
        # `cryptsymkey` and `group_id` in there.
        req = sql.join(servicegroups_table, usergroups_table,
                       ServiceGroup.group_id==UserGroup.group_id) \
                 .join(users_table, User.id==UserGroup.user_id) \
                 .select(use_labels=True) \
                 .where(User.id==self.myself_id) \
                 .where(ServiceGroup.service_id==s.id) \
                 .order_by(ServiceGroup.group_id)

        # Deal with group if specified..
        if group_id:
            req = req.where(ServiceGroup.group_id == group_id)
        
        res = meta.Session.execute(req)

        # Take the first one
        uciphers = list(res)
        if not uciphers:
            ugcgk = ''
            sgcsk = ''
            uggi = ''
        else:
            ucipher = uciphers[0]
            # WARN: these are table-name dependent!!
            ugcgk = ucipher.users_groups_cryptgroupkey
            sgcsk = ucipher.services_groups_cryptsymkey
            uggi = ucipher.users_groups_group_id

        groups_list = None
        # Load groups too if required
        if with_groups:
            groups_list = []
            req2 = sql.join(groups_table, servicegroups_table).select(use_labels=True).where(ServiceGroup.service_id == service_id)
            res2 = meta.Session.execute(req2)
            for grp in res2:
                groups_list.append((grp.groups_id, grp.groups_name))

        out = {'id': s.id,
               'url': s.url,
               'secret': s.secret,
               'machine_id': s.machine_id,
               'cryptgroupkey': ugcgk,
               'cryptsymkey': sgcsk,
               'group_id': uggi,
               'groups_list': groups_list,
               'parent_service_id': s.parent_service_id,
               'secret_last_modified': s.secret_last_modified,
               'metadata': s.metadata or {},
               'notes': s.notes or ''}

        return out

    def service_get_tree(self, service_id, with_groups=False):
        """Get a service tree, starting with service_id"""
        
        out = []
        while True:
            try:
                data = self._service_get_data(service_id,
                                              with_groups=with_groups)
            except VaultError, e:
                self.log_e('Show service: %(error)s', {"error": str(e)})
                return vaultMsg(False, str(e))
            
            out.append(data)

            if not data['parent_service_id']:
                break

            # Load the parent...
            service_id = data['parent_service_id']

            # check if we're not in an infinite loop!
            if service_id in [x['id'] for x in out]:
                self.log_e('Circular references of parent services, aborting.', {})
                return vaultMsg(False, "Circular references of parent services, aborting.")

        out.reverse()


        self.log_i('Service shown: %(service_id)s', {"service_id": service_id})
        return vaultMsg(True, "Here are the services", {'services': out})    

    
    def show(self, service_id, with_groups=False):
        """Get the specified service ID and return the hierarchy to connect
        to it or to show it.

        We need self.myself_id to be set for this function.
        """
        return self.service_get_tree(service_id, with_groups)


    def search(self, search_query, filters=None, verbose=False):
        """Do the search, and return the result tree.

        filters - must be a dictionary with options on which to constraint
                  results.

        """
        filter_types = ['groups', 'machines', 'customers']
        # Load objects on which to restrict the query:
        newfilters = {}
        if filters:
            if not isinstance(filters, dict):
                self.log_e('Search: filters must be a dictionary')
                return vaultMsg(False, "filters must be a dictionary")
            for flt in filter_types:
                # Skip filters that aren't specified
                if flt not in filters:
                    continue

                # Skip filters that are empty or None
                if not filters[flt]:
                    continue

                try:
                    newfilters[flt] = model.get_objects_ids(filters[flt], flt)
                except ValueError, e:
                    self.log_e('Search error: %(error)s' % {"error": str(e)})
                    return vaultMsg(False, str(e))

        search = model.search_query(search_query, newfilters, verbose)


        # Quick helper funcs, to create the hierarchical 'out' structure.
        def set_customer(out, c):
            "Subfunc of search"
            if out.has_key(str(c.customers_id)):
                return
            out[str(c.customers_id)] = {'name': c.customers_name,
                                        'machines': {}}
            
        def set_machine(subout, m):
            "Subfunc of search"
            if subout.has_key(str(m.machines_id)):
                return
            subout[str(m.machines_id)] = {'name': m.machines_name,
                            'fqdn': m.machines_fqdn or '',
                            'ip': m.machines_ip or '',
                            'location': m.machines_location or '',
                            'notes': m.machines_notes or '',
                            'services': {}}
            
        def set_service(subsubout, s):
            "Subfunc of search"
            subsubout[str(s.services_id)] = {'url': s.services_url,
                         'parent_service_id': s.services_parent_service_id \
                                              or '',
                         'metadata': s.services_metadata or '',
                         'notes': s.services_notes or ''}

        out = {}
        # Loop services, setup machines and customers first.
        for x in search:
            # Setup customer dans le out, pour le service
            set_customer(out, x)
            set_machine(out[str(x.customers_id)]['machines'], x)
            set_service(out[str(x.customers_id)]['machines'] \
                            [str(x.machines_id)]['services'], x)


        # Return 'out', in a nicely structured hierarchical form.
        #self.log_i('Search successfull for: %(search)s',
        #            {'search': search_query})
        return vaultMsg(True, "Here are the search results", {'results': out})


    def customer_get(self, customer_id):
        """Get a single customer's data"""
        try:
            cust = query(Customer).filter_by(id=customer_id).one()
        except InvalidReq, e:
            self.log_i('Customer not found: %(error)s', {"error": str(e)})
            return vaultMsg(False, "Customer not found: %s" % str(e))

        out = {'id': cust.id,
               'name': cust.name}

        return vaultMsg(True, "Here is the customer", {'customer': out})

    def customer_put(self, customer_id, data):
        """Put a single customer's data back to the Vault"""
        try:
            cust = query(Customer).filter_by(id=customer_id).one()
        except InvalidReq, e:
            self.log_w('Customer not found: %(customer_id)s',
                       {"customer_id": customer_id})
            return vaultMsg(False, "Customer not found: %s" % str(e))

        if 'name' in data:
            cust.name = data['name']

        meta.Session.commit()

        self.log_i('Customer c#%(customer_id)s saved successfully)s',
                  {"customer_id": customer_id})
        return vaultMsg(True, "Customer c#%s saved successfully" % customer_id)


    def customer_add(self, customer_name):
        """Add a new customer to the database"""
        nc = Customer()
        nc.name = customer_name
        nc.created_time = datetime.now()
        nc.created_user = self.myself_username
        if not self.myself_username:
            log.warning("You should *always* set myself_username on the " + \
                        self.__class__.__name__ + " object when calling "\
                        "add_customer() to log creation time and user infos.")

        meta.Session.save(nc)
        
        meta.Session.commit()

        self.log_i('Customer add: c#%(customer_id)s', {"customer_id": nc.id})
        return vaultMsg(True, 'Customer added', {'customer_id': nc.id})


    def machine_put(self, machine_id, data):
        """Put a single machine's data back to the vault"""
        try:
            m = query(Machine).filter_by(id=machine_id).one()
        except InvalidReq, e:
            self.log_w('Machine m#%(machine_id)s saved successfully',
                   {"machine_id": machine_id})
            return vaultMsg(False, "Machine not found: %s" % str(e))

        if 'customer_id' in data:
            m.customer_id = int(data['customer_id'])

        for x in ['ip', 'name', 'fqdn', 'location', 'notes']:
            if x in data:
                m.__setattr__(x, data[x])
        meta.Session.commit()

        self.log_i('Machine m#%(machine_id)s saved successfully',
                   {"machine_id": machine_id})
        return vaultMsg(True, "Machine m#%s saved successfully" % machine_id)

    def machine_get(self, machine_id):
        """Get a single machine's data"""
        try:
            m = query(Machine).filter_by(id=machine_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Machine not found: %s" % str(e))


        out = {'id': m.id,
               'ip': m.ip or '',
               'fqdn': m.fqdn or '',
               'name': m.name or '',
               'location': m.location or '',
               'notes': m.notes or '',
               'customer_id': m.customer_id}

        return vaultMsg(True, "Here is the machine", {'machine': out})  


    def machine_add(self, customer_id, name, fqdn, ip, location, notes):
        """Add a new machine to the database"""
        
        nm = Machine()
        nm.customer_id = int(customer_id)
        nm.created_time = datetime.now()
        if not name:
            return vaultMsg(False, "Missing required argument: name")
        nm.name = name
        nm.fqdn = fqdn or ''
        nm.ip = ip or ''
        nm.location = location or ''
        nm.notes = notes or ''

        meta.Session.save(nm)
        
        meta.Session.commit()

        self.log_i('Machine added: m#%(machine_id)s', {"machine_id": nm.id})
        return vaultMsg(True, "Machine added.", {'machine_id': nm.id})


    def service_add(self, machine_id, parent_service_id, url,
                    group_ids, secret, notes, metadata):
        # Get groups
        try:
            groups, group_ids = model.get_objects_list(group_ids, 'groups',
                                                       return_objects=True)
        except ValueError, e:
            return vaultMsg(False, str(e))
        
        # TODO: centralise the grant and users matching service resolution.

        # Make sure the parent service exists, if specified
        if parent_service_id:
            try:
                parent = query(Service).get(parent_service_id)
            except:
                return vaultMsg(False, "No such parent service ID.",
                                {'parent_service_id': parent_service_id})

        # seckey is the AES256 symmetric key, to be encrypted for each user.
        (seckey, ciphertext) = encrypt_secret(secret)

        # Add service effectively..
        ns = Service()
        ns.machine_id = int(machine_id)
        ns.parent_service_id = parent_service_id or None
        ns.url = url
        ns.secret = ciphertext
        ns.secret_last_modified = datetime.now()
        ns.notes = notes
        ns.metadata = metadata

        meta.Session.save(ns)

        # Encrypt symkey for each group, using it's own ElGamal pubkey
        for g in groups:
            eg = g.elgamal()
            
            nsg = ServiceGroup()
            nsg.group_id = g.id
            nsg.cryptsymkey = encrypt_longmsg(eg, seckey)
            
            ns.groups_assoc.append(nsg)

        del(seckey)

        meta.Session.commit()

        grouplist = [g.name for g in groups]

        return vaultMsg(True, "Service added.", {'service_id': ns.id,
                                                 'encrypted_for': grouplist})

    def group_get(self, group_id):
        """Get a single group's data"""
        try:
            grp = query(Group).filter_by(id=group_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Group not found: %s" % str(e))

        ug = query(UserGroup).filter_by(user_id=self.myself_id,
                                        group_id=group_id).first()

        # TODO: don't allow if grp.hidden == True and user isn't part of the
        # group.

        out = {'id': grp.id,
               'name': grp.name,
               'hidden': grp.hidden}

        if grp.hidden and (not ug or not ug.is_admin):
            return vaultMsg(False, "Group not found*")

        if ug:
            out['cryptgroupkey'] = ug.cryptgroupkey
        
        return vaultMsg(True, "Here is the group", {'group': out})


    def group_put(self, group_id, data):
        """Put a single group's data back to the Vault"""
        try:
            grp = query(Group).filter_by(id=group_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Group not found: %s" % str(e))

        ug = query(UserGroup).filter_by(user_id=self.myself_id,
                                        group_id=group_id).first()
        if not ug:
            return vaultMsg(False, "Cannot write to group: not member of group")

        if 'name' in data:
            grp.name = data['name']
            
        if 'hidden' in data:
            newhidden = bool(data['hidden'])
            # TODO: these checks must go in the XML-RPC controller.
            if not ug.is_admin and newhidden:
                return vaultMsg(False, "You need to be admin on this group "
                                "to hide the group")
            grp.hidden = newhidden

        meta.Session.commit()

        return vaultMsg(True, "Group g#%s saved successfully" % group_id)


    def group_add(self, group_name, hidden=False):
        """Add a new group the database. Nothing will be added to it
        by default"""

        # Get my User, to get my pubkey.
        me = query(User).get(self.myself_id)
        myeg = me.elgamal()
        
        # Generate keypair
        newkeys = generate_elgamal_keypair()
        
        ng = Group()
        ng.name = group_name
        ng.hidden = hidden
        ng.pubkey = serial_elgamal_pubkey(elgamal_pubkey(newkeys))

        meta.Session.save(ng)

        meta.Session.commit()

        # Add myself to the group and all other global admins.
        admins = set(query(User).filter_by(is_admin=True).all())
        admins.add(me)

        for usr in list(admins):
            nug = UserGroup()
            # Make sure I'm admin of my newly created group
            if usr == me:
                nug.is_admin = True
            nug.user_id = usr.id
            nug.cryptgroupkey = encrypt_longmsg(usr.elgamal(),
                                                serial_elgamal_privkey(
                                                    elgamal_bothkeys(newkeys)))
            ng.users_assoc.append(nug)
        
        meta.Session.commit()

        return vaultMsg(True, "Added group '%s'" % ng.name,
                        {'name': ng.name, 'group_id': int(ng.id),
                         'cryptgroupkey': nug.cryptgroupkey})

    def group_del(self, group_id):
        """Remove a group from the vault. Only if no services are associated
        with it anymore."""        
        grp = query(Group).options(eagerload('services_assoc')).filter_by(id=int(group_id)).first()

        if grp is None:
            return vaultMsg(False, "Group not found: %s" % (group_id,))
        
        if len(grp.services_assoc):
            return vaultMsg(False, "Group not empty, cannot delete")

        # Delete UserGroup elements...
        q1 = usergroups_table.delete(UserGroup.group_id==grp.id)
        meta.Session.execute(q1)

        name = grp.name
        # Delete Group and commit..
        meta.Session.delete(grp)
        meta.Session.commit()

        retval = {'name': name,
                  'group_id': group_id}
        return vaultMsg(True, 'Removed group "%s" successfully' % name, retval)


    def group_list(self, show_hidden=False, list_users=False):
        """Return a simple list of the available groups"""
        groups = query(Group).options(eagerload_all('users_assoc.user')).all()

        me = query(User).get(self.myself_id)

        out = []
        for grp in groups:
            myug = [ug for ug in grp.users_assoc if ug.user_id == me.id]
            
            res = {'id': grp.id,
                   'name': grp.name,
                   'member': me in grp.users,
                   'hidden': False,
                   'admin': False}

            if grp.hidden:
                # Only for global-admin or members, if we're asked to hide
                # hidden groups
                if not show_hidden and not me.is_admin and not myug:
                    continue
                res['hidden'] = True

            if len(myug):
                res['cryptgroupkey'] = myug[0].cryptgroupkey
                if myug[0].is_admin:
                    res['admin'] = True
            
            print "A" * 1000
            print grp.users_assoc
            print grp.users
            res['members'] = [(u.user_id, u.user.username, u.is_admin)
                              for u in grp.users_assoc]

            out.append(res)

        return vaultMsg(True, 'Here is the list of groups', {'list': out})


    def group_add_service(self, group_id, service_id, symkey):
        """Add a service to a group.

        Call servie_get() first to get the information and decrypt the symkey
        on your side, then call this function to store the symkey.

        The server-side Vault will encrypt it for the given group.
        """

        try:
            grp = query(Group).filter_by(id=group_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Group not found: %s" % str(e))

        grpeg = grp.elgamal()

        sg = query(ServiceGroup).filter_by(group_id=group_id,
                                           service_id=service_id).all()
        if len(sg):
            return vaultMsg(False, "Service is already in group")

        # Store the data in the vault.
        nsg = ServiceGroup()
        nsg.group_id = group_id
        nsg.service_id = service_id
        nsg.cryptsymkey = encrypt_longmsg(grpeg, symkey)

        meta.Session.save(nsg)
        meta.Session.commit()
        
        return vaultMsg(True, "Added service to group successfully", {})


    def group_del_service(self, group_id, service_id):
        """Remove the association between a group and a service, simply."""
        grp = query(Group).filter_by(id=group_id).first()

        if not grp:
            return vaultMsg(False, "Group not found: %s" % str(e))

        # TODO: DRY out this place, much copy from del_user and stuff
        sgs = query(ServiceGroup).filter_by(service_id=service_id).all()

        if grp.id not in [sg.group_id for sg in sgs]:
            return vaultMsg(False, "Service is not in group: %s" % str(e))

        sg = [sg for sg in sgs if grp.id == sg.group_id][0]
            
        # Make sure we don't lose all of the service's crypted information.
        if len(sgs) < 2:
            return vaultMsg(False, "This is the last group this service is in. Either delete the service, or add it to another group first")

        # Remove the GroupService from the Group object.
        meta.Session.delete(sg)
        meta.Session.commit()

        return vaultMsg(True, "Removed service from group successfully")


    def group_add_user(self, group_id, user, is_admin=False,
                       cryptgroupkey=None):
        """Add a user to a group. Call once to retrieve information,
        and a second time with cryptgroupkey to save cipher information.

        The second call should give the group's privkey, encrypted by the
        remote user for the user being added.

        is_admin - Gives admin privileges to the user being added or not.
        user - User can be a username or a user_id
        """
        try:
            grp = query(Group).filter_by(id=group_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Group not found: %s" % str(e))

        # Verify if I'm admin on that group
        ug = query(UserGroup).filter_by(group_id=group_id,
                                        user_id=self.myself_id).first()

        # Make sure I'm in that group (to be able to decrypt the groupkey)
        if not ug:
            return vaultMsg(False, "You are not part of that group")

        # No admin checks, you NEED to be in the group to return consistantly
        # encrypted cryptgroupkey

        # Find added user
        try:
            usr = get_user(user)
        except LookupError, e:
            return vaultMsg(False, "User %s not found: %s" % (user, str(e)))


        # Make sure we don't double the group access.
        if query(UserGroup).filter_by(group_id=group_id,
                                  user_id=usr.id).first():
            return vaultMsg(False, "User %s is already in that group" % user)
        
        if not cryptgroupkey:
            # Return the required information for a second call to work.
            ret = {'user_id': int(usr.id),
                   'group_id': int(group_id),
                   'userpubkey': usr.pubkey,
                   'cryptgroupkey': ug.cryptgroupkey}
            return vaultMsg(True, "Continue, send the cryptgroupkey back, encrypted by the userpubkey provided", ret)
        
        nug = UserGroup()
        nug.user_id = usr.id
        nug.group_id = group_id
        nug.is_admin = is_admin
        # NOTE: This command trusts that the remote user adding another user
        # enters valid data as a cryptgroupkey. If junk is added in that field,
        # the added user simply won't be able to access the group's data.
        # Also will he need to be removed and added successfully to the group
        # for everything to work.
        nug.cryptgroupkey = cryptgroupkey

        meta.Session.save(nug)
        meta.Session.commit()
        
        return vaultMsg(True, "Added user to group successfully")


    def group_del_user(self, group_id, user):
        """Remove the association between a group and a user.

        Make sure there are still 'is_admin' users associated with the group
        to be able to give access to someone else in case of need.

        SFLvault will refuse to delete a user from a group if no other users
        are part of that group, and if the group still has some services
        associated with it. First, call group_del_service from the group for
        all services, then delete the users, and then the group.

        user - can be a username or a user_id
        """
        # TODO: DRY out this place, much copy from group_add_user
        try:
            grp = query(Group).filter_by(id=group_id).one()
        except InvalidReq, e:
            return vaultMsg(False, "Group not found: %s" % str(e))

        # Verify if I'm admin on that group
        ugs = query(UserGroup).filter_by(group_id=group_id).all()

        myug = [ug for ug in ugs if ug.user_id == self.myself_id]
        if not myug:
            return vaultMsg(False, "You are not part of that group")

        ug = myug[0] # This is my UserGroup row.
        
        me = query(User).get(self.myself_id)
        
        if not ug.is_admin and not me.is_admin:
            return vaultMsg(False, "You are not admin on that group (nor global admin)")

        # Find added user
        try:
            usr = get_user(user)
        except LookupError, e:
            return vaultMsg(False, "User %s not found: %s" % (user, str(e)))

        hisug = [ug for ug in ugs if ug.user_id == usr.id]

        if not hisug:
            return vaultMsg(False, "User isn't part of the group")

        # Check I'm not removing myself
        if self.myself_id == usr.id:
            return vaultMsg(False, "Cannot remove yourself from a group")

        # This also prevents from removing myself from the group.
        if len(ugs) < 2:
            return vaultMsg(False, "Only one user left! Cannot leave a group unattended, otherwise all service will become lost!")

        # If we're using the library, we must make sure there is still a
        # group admin left.  This simplifiest the rescues :)
        id_admins = [x.id for x in ugs if x.is_admin]
        if id_admins == [usr.id]:
            return vaultMsg(False, "Each group must have at least one group-admin.  You cannot delete the last group-admin.")
        
        ohoh = ''
        if not id_admins:
            ohoh = " - WARNING: there are no more group-admins in this group.  Ask a global-admin to elect someone group-admin for further management of this group."

        meta.Session.delete(hisug[0])
        meta.Session.commit()

        return vaultMsg(True, "Removed user from group successfully" + ohoh, {})
        

    def customer_del(self, customer_id):
        """Delete a customer from database, bringing along all it's machines
        and services
        """
        # Get customer
        cust = query(model.Customer).options(model.eagerload('machines'))\
                                    .get(int(customer_id))

        if not cust:
            return vaultMsg(True, "No such customer: c#%s" % customer_id)

        # Get all the services that will be deleted
        servs = query(model.Service).join(['machine', 'customer']) \
                     .filter(model.Customer.id == customer_id) \
                     .all()
        servs_ids = [s.id for s in servs]

        # Make sure no service is child of this one
        childs = query(model.Service) \
                     .filter(model.Service.parent_service_id.in_(servs_ids))\
                     .all()

        # Don't bother for parents/childs if we're going to delete it anyway.
        remnants = list(set(childs).difference(set(servs)))

        if len(remnants):
            # There are still some childs left, we can't delete this one.
            retval = []
            for x in remnants:
                retval.append({'id': x.id, 'url': x.url})
                
            return vaultMsg(False, "Services still child of this customer's machine's services",
                            {'childs': retval})

        # Delete all related groupciphers
        d = sql.delete(model.servicegroups_table) \
               .where(model.servicegroups_table.c.service_id.in_(servs_ids))
        # Delete the services related to customer_id's machines
        d2 = sql.delete(model.services_table) \
                .where(model.services_table.c.id.in_(servs_ids))
        # Delete the machines related to customer_id
        mach_ids = [m.id for m in cust.machines]
        d3 = sql.delete(model.machines_table) \
                .where(model.machines_table.c.id.in_(mach_ids))
        # Delete the customer
        d4 = sql.delete(model.customers_table) \
                .where(model.customers_table.c.id == customer_id)

        meta.Session.execute(d)
        meta.Session.execute(d2)
        meta.Session.execute(d3)
        meta.Session.execute(d4)
        
        meta.Session.commit()

        return vaultMsg(True,
                        'Deleted customer c#%s successfully' % customer_id)


    def machine_del(self, machine_id):
        """Delete a machine from database, bringing on all child services."""
        
        # Get machine
        machine = query(model.Machine).get(int(machine_id))

        if not machine:
            return vaultMsg(True, "No such machine: m#%s" % machine_id)

        # Get all the services that will be deleted
        servs = query(model.Service).join('machine') \
                     .filter(model.Machine.id == machine_id).all()
        servs_ids = [s.id for s in servs]

        # Make sure no service is child of this one
        childs = query(model.Service) \
                     .filter(model.Service.parent_service_id.in_(servs_ids))\
                     .all()

        # Don't bother for parents/childs if we're going to delete it anyway.
        remnants = list(set(childs).difference(set(servs)))

        if len(remnants):
            # There are still some childs left, we can't delete this one.
            retval = []
            for x in remnants:
                retval.append({'id': x.id, 'url': x.url})
                
            return vaultMsg(False, "Services still child of this machine's services",
                            {'childs': retval})


        # Delete all related groupciphers
        d = sql.delete(model.servicegroups_table) \
               .where(model.servicegroups_table.c.service_id.in_(servs_ids))
        # Delete the services related to machine_id
        d2 = sql.delete(model.services_table) \
                .where(model.services_table.c.id.in_(servs_ids))
        # Delete the machine
        d3 = sql.delete(model.machines_table) \
                .where(model.machines_table.c.id == machine_id)

        meta.Session.execute(d)
        meta.Session.execute(d2)
        meta.Session.execute(d3)
        
        meta.Session.commit()

        return vaultMsg(True, 'Deleted machine m#%s successfully' % machine_id)


    def service_del(self, service_id):
        """Delete a service, making sure no other child remains attached."""
        # Integerize
        service_id = int(service_id)
        # Get service
        serv = query(model.Service).get(int(service_id))

        if not serv:
            return vaultMsg(True, "No such service: s#%s" % service_id)

        # Make sure no service is child of this one
        childs = query(model.Service).filter_by(parent_service_id=service_id).all()
        if len(childs):
            # There are still some childs left, we can't delete this one.
            retval = []
            for x in childs:
                retval.append({'id': x.id, 'url': x.url})
                
            return vaultMsg(False, 'Services still child of this service',
                            {'childs': retval})

        # TODO: verify permissions:
        #  if either user is global admin, or is admin of all groups in which this
        #  service is in, otherwise, disallow.
        
        # Delete all related user-ciphers
        d = sql.delete(model.servicegroups_table) \
               .where(model.servicegroups_table.c.service_id == service_id)
        # Delete the service
        d2 = sql.delete(services_table) \
                .where(services_table.c.id == service_id)
        
        meta.Session.execute(d)
        meta.Session.execute(d2)
        meta.Session.commit()

        return vaultMsg(True, 'Deleted service s#%s successfully' % service_id)



    def customer_list(self):
        lst = query(Customer).all()

        out = []
        for x in lst:
            nx = {'id': x.id, 'name': x.name}
            out.append(nx)

        return vaultMsg(True, 'Here is the customer list', {'list': out})



    def machine_list(self, customer_id=None):
        """Return a simple list of the machines"""    
        sel = sql.join(customers_table, machines_table) \
                 .select(use_labels=True) \
                 .order_by(Customer.id)

        # Filter also..
        if customer_id:
            sel = sel.where(Customer.id==customer_id)
        
        lst = meta.Session.execute(sel)

        out = [{'id': x.machines_id, 'name': x.machines_name,
                  'fqdn': x.machines_fqdn, 'ip': x.machines_ip,
                  'location': x.machines_location, 'notes': x.machines_notes,
                  'customer_id': x.customers_id,
                  'customer_name': x.customers_name}
               for x in lst]

        return vaultMsg(True, "Here is the machines list", {'list': out})
    

    def service_list(self, machine_id=None, customer_id=None):
        """Return a simple list of the services"""
        services = query(Service).join('machine')

        # Filter also..
        if machine_id:
            services = services.filter(Service.machine_id==int(machine_id))
        if customer_id:
            services = services.filter(Machine.customer_id==int(customer_id))

        out = [{'id': s.id, 'url': s.url,
                   'parent_service_id': s.parent_service_id,
                   'metadata': s.metadata, 'notes': s.notes,
                   'machine_id': s.machine_id,
                   'secret': s.secret,
                   'secret_last_modified': s.secret_last_modified,
                }
               for s in services]

        return vaultMsg(True, "Here is the machines list", {'list': out})


    def service_passwd(self, service_id, newsecret):
        """Change the passwd for a given service"""
        # number please
        service_id = int(service_id)

        serv = query(Service).get(service_id)
        groups = serv.groups

        (seckey, ciphertext) = encrypt_secret(newsecret)
        serv.secret = ciphertext
        
        # TODO absolutely:  verify this requesting user has access to the
        # password first.  YES, he can add a new password, but not modify
        # something he doesn't have access to, first.

        for sg in serv.groups_assoc:
            eg = [g for g in groups if g.id == sg.group_id][0].elgamal()
            sg.cryptsymkey = encrypt_longmsg(eg, seckey)

        meta.Session.commit()

        grouplist = [g.name for g in groups]
        
        return vaultMsg(True, "Service added.", {'service_id': service_id,
                                                 'encrypted_for': grouplist})



try:
    from setuptools import setup, find_packages
except ImportError:
    from distribute_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages



setup(
    name='SFLvault-server',
    version="0.7.8.1",
    description='Networked credentials store and authentication manager - Server',
    author='Alexandre Bourget',
    author_email='alexandre.bourget@savoirfairelinux.com',
    url='http://www.sflvault.org',
    license='GPLv3',
    install_requires=["Pylons==0.9.7",
                      "SQLAlchemy==0.5.8",
                      "pysqlite",
                      "simplejson",
                      "SFLvault-common==0.7.8.1",
                      "Routes==1.10.3",
                      "Paste==1.7.3.1",
                      ],
    # For server installation:
    #  "ipython"
    #  "pysqlite"
    # For development installation:
    #  "nosexml"
    #  "elementtree"
    #  "coverage"
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    package_data={'sflvault': ['i18n/*/LC_MESSAGES/*.mo']},
    #message_extractors = {'sflvault': [
    #        ('**.py', 'python', None),
    #        ('templates/**.mako', 'mako', None),
    #        ('public/**', 'ignore', None)]},
    entry_points="""
    [paste.app_factory]
    main = sflvault.config.middleware:make_app

    [paste.app_install]
    main = pylons.util:PylonsInstaller
    """,
)


